
''' A simple starting point for curver. '''

import curver

def main():
    ''' Describe how to start curver. '''
    
    print('curver %s' % curver.__version__)
    print('Some basic commands:')
    print('  > python -m curver.app      # To start the curver GUI.')
    print('Or start Python and import curver directly.')

if __name__ == '__main__':
    main()

